// --------------------------------------------------------------------
// Copyright (c) 2005 by Terasic Technologies Inc. 
// --------------------------------------------------------------------
//
// Permission:
//
//   Terasic grants permission to use and modify this code for use
//   in synthesis for all Terasic Development Boards and Altera Development 
//   Kits made by Terasic.  Other use of this code, including the selling 
//   ,duplication, or modification of any portion is strictly prohibited.
//
// Disclaimer:
//
//   This VHDL/Verilog or C/C++ source code is intended as a design reference
//   which illustrates how these types of functions can be implemented.
//   It is the user's responsibility to verify their design for
//   consistency and functionality through the use of formal
//   verification methods.  Terasic provides no warranty regarding the use 
//   or functionality of this code.
//
// --------------------------------------------------------------------
//           
//                     Terasic Technologies Inc
//                     356 Fu-Shin E. Rd Sec. 1. JhuBei City,
//                     HsinChu County, Taiwan
//                     302
//
//                     web: http://www.terasic.com/
//                     email: support@terasic.com
//
// --------------------------------------------------------------------
// Major Functions:	DE2 TV Box
// --------------------------------------------------------------------
//
// Revision History :
// --------------------------------------------------------------------
//   Ver  :| Author            :| Mod. Date :| Changes Made:
//   V1.0 :| Joe Yang	       :| 05/07/05  :| Initial Revision
//   V1.1 :| Johnny Chen       :| 05/09/05  :| Changed YCbCr2RGB Block,
//											   RGB output 8 Bits => 10 Bits
//   V1.2 :| Johnny Chen	   :| 05/10/05  :| H_SYNC & V_SYNC Timing fixed.
//   V1.3 :| Johnny Chen       :| 05/11/16  :| Added FLASH Address FL_ADDR[21:20]
//   V1.4 :| Joe Yang	       :| 06/07/20  :| Modify Output Color
//	 V2.0 :| Johnny Chen	   :| 06/11/20	:| New Version for DE2 v2.X PCB.
// --------------------------------------------------------------------

module DE2_TV
	(
		////////////////////	Clock Input	 	////////////////////	 
		OSC_27,							//	27 MHz
		OSC_50,							//	50 MHz
		EXT_CLOCK,						//	External Clock
		////////////////////	Push Button		////////////////////
		KEY,							//	Button[3:0]
		////////////////////	DPDT Switch		////////////////////
		DPDT_SW,						//	DPDT Switch[17:0]
		////////////////////	7-SEG Dispaly	////////////////////
		HEX0,							//	Seven Segment Digital 0
		HEX1,							//	Seven Segment Digital 1
		HEX2,							//	Seven Segment Digital 2
		HEX3,							//	Seven Segment Digital 3
		HEX4,							//	Seven Segment Digital 4
		HEX5,							//	Seven Segment Digital 5
		HEX6,							//	Seven Segment Digital 6
		HEX7,							//	Seven Segment Digital 7
		////////////////////////	LED		////////////////////////
		LED_GREEN,						//	LED Green[8:0]
		LED_RED,						//	LED Red[17:0]
		////////////////////////	UART	////////////////////////
		UART_TXD,						//	UART Transmitter
		UART_RXD,						//	UART Receiver
		////////////////////////	IRDA	////////////////////////
		IRDA_TXD,						//	IRDA Transmitter
		IRDA_RXD,						//	IRDA Receiver
		/////////////////////	SDRAM Interface		////////////////
		DRAM_DQ,						//	SDRAM Data bus 16 Bits
		DRAM_ADDR,						//	SDRAM Address bus 12 Bits
		DRAM_LDQM,						//	SDRAM Low-byte Data Mask 
		DRAM_UDQM,						//	SDRAM High-byte Data Mask
		DRAM_WE_N,						//	SDRAM Write Enable
		DRAM_CAS_N,						//	SDRAM Column Address Strobe
		DRAM_RAS_N,						//	SDRAM Row Address Strobe
		DRAM_CS_N,						//	SDRAM Chip Select
		DRAM_BA_0,						//	SDRAM Bank Address 0
		DRAM_BA_1,						//	SDRAM Bank Address 0
		DRAM_CLK,						//	SDRAM Clock
		DRAM_CKE,						//	SDRAM Clock Enable
		////////////////////	Flash Interface		////////////////
		FL_DQ,							//	FLASH Data bus 8 Bits
		FL_ADDR,						//	FLASH Address bus 22 Bits
		FL_WE_N,						//	FLASH Write Enable
		FL_RST_N,						//	FLASH Reset
		FL_OE_N,						//	FLASH Output Enable
		FL_CE_N,						//	FLASH Chip Enable
		////////////////////	SRAM Interface		////////////////
		SRAM_DQ,						//	SRAM Data bus 16 Bits
		SRAM_ADDR,						//	SRAM Adress bus 18 Bits
		SRAM_UB_N,						//	SRAM High-byte Data Mask 
		SRAM_LB_N,						//	SRAM Low-byte Data Mask 
		SRAM_WE_N,						//	SRAM Write Enable
		SRAM_CE_N,						//	SRAM Chip Enable
		SRAM_OE_N,						//	SRAM Output Enable
		////////////////////	ISP1362 Interface	////////////////
		OTG_DATA,						//	ISP1362 Data bus 16 Bits
		OTG_ADDR,						//	ISP1362 Address 2 Bits
		OTG_CS_N,						//	ISP1362 Chip Select
		OTG_RD_N,						//	ISP1362 Write
		OTG_WR_N,						//	ISP1362 Read
		OTG_RST_N,						//	ISP1362 Reset
		OTG_FSPEED,						//	USB Full Speed,	0 = Enable, Z = Disable
		OTG_LSPEED,						//	USB Low Speed, 	0 = Enable, Z = Disable
		OTG_INT0,						//	ISP1362 Interrupt 0
		OTG_INT1,						//	ISP1362 Interrupt 1
		OTG_DREQ0,						//	ISP1362 DMA Request 0
		OTG_DREQ1,						//	ISP1362 DMA Request 1
		OTG_DACK0_N,					//	ISP1362 DMA Acknowledge 0
		OTG_DACK1_N,					//	ISP1362 DMA Acknowledge 1
		////////////////////	LCD Module 16X2		////////////////
		LCD_ON,							//	LCD Power ON/OFF
		LCD_BLON,						//	LCD Back Light ON/OFF
		LCD_RW,							//	LCD Read/Write Select, 0 = Write, 1 = Read
		LCD_EN,							//	LCD Enable
		LCD_RS,							//	LCD Command/Data Select, 0 = Command, 1 = Data
		LCD_DATA,						//	LCD Data bus 8 bits
		////////////////////	SD_Card Interface	////////////////
		SD_DAT,							//	SD Card Data
		SD_DAT3,						//	SD Card Data 3
		SD_CMD,							//	SD Card Command Signal
		SD_CLK,							//	SD Card Clock
		////////////////////	USB JTAG link	////////////////////
		TDI,  							// CPLD -> FPGA (data in)
		TCK,  							// CPLD -> FPGA (clk)
		TCS,  							// CPLD -> FPGA (CS)
	    TDO,  							// FPGA -> CPLD (data out)
		////////////////////	I2C		////////////////////////////
		I2C_SDAT,						//	I2C Data
		I2C_SCLK,						//	I2C Clock
		////////////////////	PS2		////////////////////////////
		PS2_DAT,						//	PS2 Data
		PS2_CLK,						//	PS2 Clock
		////////////////////	VGA		////////////////////////////
		VGA_CLK,   						//	VGA Clock
		VGA_HS,							//	VGA H_SYNC
		VGA_VS,							//	VGA V_SYNC
		VGA_BLANK,						//	VGA BLANK
		VGA_SYNC,						//	VGA SYNC
		VGA_R,   						//	VGA Red[9:0]
		VGA_G,	 						//	VGA Green[9:0]
		VGA_B,  						//	VGA Blue[9:0]
		////////////	Ethernet Interface	////////////////////////
		ENET_DATA,						//	DM9000A DATA bus 16Bits
		ENET_CMD,						//	DM9000A Command/Data Select, 0 = Command, 1 = Data
		ENET_CS_N,						//	DM9000A Chip Select
		ENET_WR_N,						//	DM9000A Write
		ENET_RD_N,						//	DM9000A Read
		ENET_RST_N,						//	DM9000A Reset
		ENET_INT,						//	DM9000A Interrupt
		ENET_CLK,						//	DM9000A Clock 25 MHz
		////////////////	Audio CODEC		////////////////////////
		AUD_ADCLRCK,					//	Audio CODEC ADC LR Clock
		AUD_ADCDAT,						//	Audio CODEC ADC Data
		AUD_DACLRCK,					//	Audio CODEC DAC LR Clock
		AUD_DACDAT,						//	Audio CODEC DAC Data
		AUD_BCLK,						//	Audio CODEC Bit-Stream Clock
		AUD_XCK,						//	Audio CODEC Chip Clock
		////////////////	TV Decoder		////////////////////////
		TD_DATA,    					//	TV Decoder Data bus 8 bits
		TD_HS,							//	TV Decoder H_SYNC
		TD_VS,							//	TV Decoder V_SYNC
		TD_RESET,						//	TV Decoder Reset
		TD_CLK,							//	TV Decoder Line Locked Clock
		////////////////////	GPIO	////////////////////////////
		GPIO_0,							//	GPIO Connection 0
		GPIO_1							//	GPIO Connection 1
	);

////////////////////////	Clock Input	 	////////////////////////
input			OSC_27;					//	27 MHz
input			OSC_50;					//	50 MHz
input			EXT_CLOCK;				//	External Clock
////////////////////////	Push Button		////////////////////////
input	[3:0]	KEY;					//	Button[3:0]
////////////////////////	DPDT Switch		////////////////////////
input	[17:0]	DPDT_SW;				//	DPDT Switch[17:0]
////////////////////////	7-SEG Dispaly	////////////////////////
output	[6:0]	HEX0;					//	Seven Segment Digital 0
output	[6:0]	HEX1;					//	Seven Segment Digital 1
output	[6:0]	HEX2;					//	Seven Segment Digital 2
output	[6:0]	HEX3;					//	Seven Segment Digital 3
output	[6:0]	HEX4;					//	Seven Segment Digital 4
output	[6:0]	HEX5;					//	Seven Segment Digital 5
output	[6:0]	HEX6;					//	Seven Segment Digital 6
output	[6:0]	HEX7;					//	Seven Segment Digital 7
////////////////////////////	LED		////////////////////////////
output	[8:0]	LED_GREEN;				//	LED Green[8:0]
output	[17:0]	LED_RED;				//	LED Red[17:0]
////////////////////////////	UART	////////////////////////////
output			UART_TXD;				//	UART Transmitter
input			UART_RXD;				//	UART Receiver
////////////////////////////	IRDA	////////////////////////////
output			IRDA_TXD;				//	IRDA Transmitter
input			IRDA_RXD;				//	IRDA Receiver
///////////////////////		SDRAM Interface	////////////////////////
inout	[15:0]	DRAM_DQ;				//	SDRAM Data bus 16 Bits
output	[11:0]	DRAM_ADDR;				//	SDRAM Address bus 12 Bits
output			DRAM_LDQM;				//	SDRAM Low-byte Data Mask 
output			DRAM_UDQM;				//	SDRAM High-byte Data Mask
output			DRAM_WE_N;				//	SDRAM Write Enable
output			DRAM_CAS_N;				//	SDRAM Column Address Strobe
output			DRAM_RAS_N;				//	SDRAM Row Address Strobe
output			DRAM_CS_N;				//	SDRAM Chip Select
output			DRAM_BA_0;				//	SDRAM Bank Address 0
output			DRAM_BA_1;				//	SDRAM Bank Address 0
output			DRAM_CLK;				//	SDRAM Clock
output			DRAM_CKE;				//	SDRAM Clock Enable
////////////////////////	Flash Interface	////////////////////////
inout	[7:0]	FL_DQ;					//	FLASH Data bus 8 Bits
output	[21:0]	FL_ADDR;				//	FLASH Address bus 22 Bits
output			FL_WE_N;				//	FLASH Write Enable
output			FL_RST_N;				//	FLASH Reset
output			FL_OE_N;				//	FLASH Output Enable
output			FL_CE_N;				//	FLASH Chip Enable
////////////////////////	SRAM Interface	////////////////////////
inout	[15:0]	SRAM_DQ;				//	SRAM Data bus 16 Bits
output	[17:0]	SRAM_ADDR;				//	SRAM Address bus 18 Bits
output			SRAM_UB_N;				//	SRAM High-byte Data Mask 
output			SRAM_LB_N;				//	SRAM Low-byte Data Mask 
output			SRAM_WE_N;				//	SRAM Write Enable
output			SRAM_CE_N;				//	SRAM Chip Enable
output			SRAM_OE_N;				//	SRAM Output Enable
////////////////////	ISP1362 Interface	////////////////////////
inout	[15:0]	OTG_DATA;				//	ISP1362 Data bus 16 Bits
output	[1:0]	OTG_ADDR;				//	ISP1362 Address 2 Bits
output			OTG_CS_N;				//	ISP1362 Chip Select
output			OTG_RD_N;				//	ISP1362 Write
output			OTG_WR_N;				//	ISP1362 Read
output			OTG_RST_N;				//	ISP1362 Reset
output			OTG_FSPEED;				//	USB Full Speed,	0 = Enable, Z = Disable
output			OTG_LSPEED;				//	USB Low Speed, 	0 = Enable, Z = Disable
input			OTG_INT0;				//	ISP1362 Interrupt 0
input			OTG_INT1;				//	ISP1362 Interrupt 1
input			OTG_DREQ0;				//	ISP1362 DMA Request 0
input			OTG_DREQ1;				//	ISP1362 DMA Request 1
output			OTG_DACK0_N;			//	ISP1362 DMA Acknowledge 0
output			OTG_DACK1_N;			//	ISP1362 DMA Acknowledge 1
////////////////////	LCD Module 16X2	////////////////////////////
inout	[7:0]	LCD_DATA;				//	LCD Data bus 8 bits
output			LCD_ON;					//	LCD Power ON/OFF
output			LCD_BLON;				//	LCD Back Light ON/OFF
output			LCD_RW;					//	LCD Read/Write Select, 0 = Write, 1 = Read
output			LCD_EN;					//	LCD Enable
output			LCD_RS;					//	LCD Command/Data Select, 0 = Command, 1 = Data
////////////////////	SD Card Interface	////////////////////////
inout			SD_DAT;					//	SD Card Data
inout			SD_DAT3;				//	SD Card Data 3
inout			SD_CMD;					//	SD Card Command Signal
output			SD_CLK;					//	SD Card Clock
////////////////////////	I2C		////////////////////////////////
inout			I2C_SDAT;				//	I2C Data
output			I2C_SCLK;				//	I2C Clock
////////////////////////	PS2		////////////////////////////////
input		 	PS2_DAT;				//	PS2 Data
input			PS2_CLK;				//	PS2 Clock
////////////////////	USB JTAG link	////////////////////////////
input  			TDI;					// CPLD -> FPGA (data in)
input  			TCK;					// CPLD -> FPGA (clk)
input  			TCS;					// CPLD -> FPGA (CS)
output 			TDO;					// FPGA -> CPLD (data out)
////////////////////////	VGA			////////////////////////////
output			VGA_CLK;   				//	VGA Clock
output			VGA_HS;					//	VGA H_SYNC
output			VGA_VS;					//	VGA V_SYNC
output			VGA_BLANK;				//	VGA BLANK
output			VGA_SYNC;				//	VGA SYNC
output	[9:0]	VGA_R;   				//	VGA Red[9:0]
output	[9:0]	VGA_G;	 				//	VGA Green[9:0]
output	[9:0]	VGA_B;   				//	VGA Blue[9:0]
////////////////	Ethernet Interface	////////////////////////////
inout	[15:0]	ENET_DATA;				//	DM9000A DATA bus 16Bits
output			ENET_CMD;				//	DM9000A Command/Data Select, 0 = Command, 1 = Data
output			ENET_CS_N;				//	DM9000A Chip Select
output			ENET_WR_N;				//	DM9000A Write
output			ENET_RD_N;				//	DM9000A Read
output			ENET_RST_N;				//	DM9000A Reset
input			ENET_INT;				//	DM9000A Interrupt
output			ENET_CLK;				//	DM9000A Clock 25 MHz
////////////////////	Audio CODEC		////////////////////////////
inout			AUD_ADCLRCK;			//	Audio CODEC ADC LR Clock
input			AUD_ADCDAT;				//	Audio CODEC ADC Data
inout			AUD_DACLRCK;			//	Audio CODEC DAC LR Clock
output			AUD_DACDAT;				//	Audio CODEC DAC Data
inout			AUD_BCLK;				//	Audio CODEC Bit-Stream Clock
output			AUD_XCK;				//	Audio CODEC Chip Clock
////////////////////	TV Devoder		////////////////////////////
input	[7:0]	TD_DATA;    			//	TV Decoder Data bus 8 bits
input			TD_HS;					//	TV Decoder H_SYNC
input			TD_VS;					//	TV Decoder V_SYNC
output			TD_RESET;				//	TV Decoder Reset
input			TD_CLK;					//	TV Decoder Line Locked Clock
////////////////////////	GPIO	////////////////////////////////
inout	[35:0]	GPIO_0;					//	GPIO Connection 0
inout	[35:0]	GPIO_1;					//	GPIO Connection 1
////////////////////////////////////////////////////////////////////

//	Enable TV Decoder
assign	TD_RESET = KEY[0];

//	For Audio CODEC
wire	AUD_CTRL_CLK; // For Audio Controller
assign	AUD_XCK	= AUD_CTRL_CLK;

//	7 segment LUT
SEG7_LUT_8 			u0 (.oSEG0(HEX0),
						.oSEG1(HEX1),
						.oSEG2(HEX2),
						.oSEG3(HEX3),
						.oSEG4(HEX4),
						.oSEG5(HEX5),
						.oSEG6(HEX6),
						.oSEG7(HEX7),
						.iDIG({8'b0,1'b0,11'd0,1'b0,score}/*DPDT_SW*/)
						//.iDIG({8'b0,1'b0,drCenterX,1'b0,drCenterY}/*DPDT_SW*/) 
);

//	Audio CODEC and video decoder setting
I2C_AV_Config 		u1 (// Host Side
						.iCLK(OSC_50),
						.iRST_N(KEY[0]),
						// I2C Side
						.I2C_SCLK(I2C_SCLK),
						.I2C_SDAT(I2C_SDAT)
);

//	TV Decoder Stable Check
TD_Detect	 		u2 (.oTD_Stable(TD_Stable),
						.iTD_VS(TD_VS),
						.iTD_HS(TD_HS),
						.iRST_N(KEY[0])
);

//	Reset Delay Timer
Reset_Delay			u3 (.iCLK(OSC_50),
						.iRST(TD_Stable),
						.oRST_0(DLY0),
						.oRST_1(DLY1),
						.oRST_2(DLY2)
);

//	ITU-R 656 to YUV 4:2:2
ITU_656_Decoder		u4 (// TV Decoder Input
						.iTD_DATA(TD_DATA),
						// Position Output
						.oTV_X(TV_X),
						// YUV 4:2:2 Output
						.oYCbCr(YCbCr),
						.oDVAL(TV_DVAL),
						// Control Signals
						.iSwap_CbCr(Quotient[0]),
						.iSkip(Remain==4'h0),
						.iRST_N(DLY1),
						.iCLK_27(TD_CLK)
);

//	For Down Sample 720 to 640
DIV 				u5 (.aclr(!DLY0),	
						.clock(TD_CLK),
						.denom(4'h9),
						.numer(TV_X),
						.quotient(Quotient),
						.remain(Remain)
);

//	SDRAM frame buffer
Sdram_Control_4Port u6 (// HOST Side
						.REF_CLK(OSC_27),
						.CLK_18(AUD_CTRL_CLK),
						.RESET_N(1'b1),
						// FIFO Write Side 1
						.WR1_DATA(YCbCr),
						.WR1(TV_DVAL),
						.WR1_FULL(WR1_FULL),
						.WR1_ADDR(0),
						.WR1_MAX_ADDR(640*507),		//	525-18
						.WR1_LENGTH(9'h80),
						.WR1_LOAD(!DLY0),
						.WR1_CLK(TD_CLK),
						// FIFO Read Side 1
						.RD1_DATA(m1YCbCr),
						.RD1(m1VGA_Read),
						.RD1_ADDR(640*13),			//	Read odd field and bypess blanking
						.RD1_MAX_ADDR(640*253),
						.RD1_LENGTH(9'h80),
						.RD1_LOAD(!DLY0),
						.RD1_CLK(OSC_27),
						// FIFO Read Side 2
						.RD2_DATA(m2YCbCr),
						.RD2(m2VGA_Read),
						.RD2_ADDR(640*267),			//	Read even field and bypess blanking
						.RD2_MAX_ADDR(640*507),
						.RD2_LENGTH(9'h80),
						.RD2_LOAD(!DLY0),
						.RD2_CLK(OSC_27),
						// SDRAM Side
						.SA(DRAM_ADDR),
						.BA({DRAM_BA_1,DRAM_BA_0}),
						.CS_N(DRAM_CS_N),
						.CKE(DRAM_CKE),
						.RAS_N(DRAM_RAS_N),
						.CAS_N(DRAM_CAS_N),
						.WE_N(DRAM_WE_N),
						.DQ(DRAM_DQ),
						.DQM({DRAM_UDQM,DRAM_LDQM}),
						.SDR_CLK(DRAM_CLK)
);

// YUV 4:2:2 to YUV 4:4:4
YUV422_to_444 		u7 (// YUV 4:2:2 Input
						.iYCbCr(mYCbCr),
						// YUV	4:4:4 Output
						.oY(mY),
						.oCb(mCb),
						.oCr(mCr),
						// Control Signals
						.iX(VGA_X),
						.iCLK(OSC_27),
						.iRST_N(DLY0));

//	YCbCr 8-bit to RGB-10 bit 
YCbCr2RGB 			u8 (//	Output Side
						.Red(mRed),
						.Green(mGreen),
						.Blue(mBlue),
						.oDVAL(mDVAL),
						//	Input Side
						.iY(mY),
						.iCb(mCb),
						.iCr(mCr),
						.iDVAL(VGA_Read),
						//	Control Signal
						.iRESET(!DLY2),
						.iCLK(OSC_27)
);
/*
// DetectGREEN
detectColor			dr (.oRed(dgRed),
						.oGreen(dgGreen),
						.oBlue(dgBlue),
						.oDVAL(dgDVAL),
						.iColorTarget(22'h040FFF), // Color value
						.iRed(mRed),
						.iGreen(mGreen),
						.iBlue(mBlue),
						.targetRed(10'b0),
						.targetGreen(10'b1111111111),
						.targetBlue(10'b0),
						.iDVAL(mDVAL),
						/////////////////////
						.iX_Cont(VGA_X),
						.iY_Cont(VGA_Y),
						.oXLeft(dgXLeft),
						.oXRight(dgXRight),
						.oYTop(dgYTop),
						.oYBot(dgYBot),
						.oXLeftPrev(dgXLeftPrev),
						.oXRightPrev(dgXRightPrev),
						.oYTopPrev(dgYTopPrev),
						.oYBotPrev(dgYBotPrev),
						.oXLeftPrev2(dgXLeftPrev2),
						.oXRightPrev2(dgXRightPrev2),
						.oYTopPrev2(dgYTopPrev2),
						.oYBotPrev2(dgYBotPrev2),
						.oXLeftPrev3(dgXLeftPrev3),
						.oXRightPrev3(dgXRightPrev3),
						.oYTopPrev3(dgYTopPrev3),
						.oYBotPrev3(dgYBotPrev3),
						.oRedArea(dgRedArea),
						/////////////////////
						.iCLK(OSC_27),
						.iRST(!DLY2)
);
*/
// DetectRED
detectColor			dg (.oRed(drRed),
						.oGreen(drGreen),
						.oBlue(drBlue),
						.oDVAL(drDVAL),
						.iColorTarget(22'h04FC00), // Color value
						.iRed(mRed),
						.iGreen(mGreen),
						.iBlue(mBlue),
						.targetRed(10'b1111111111),
						.targetGreen(10'b0000000000),
						.targetBlue(10'b0),
						//.targetRed(targetR),
						//.targetGreen(targetG),
						//.targetBlue(targetB),
						.iDVAL(mDVAL),
						/////////////////////
						.iX_Cont(VGA_X),
						.iY_Cont(VGA_Y),
						.oXLeft(drXLeft),
						.oXRight(drXRight),
						.oYTop(drYTop),
						.oYBot(drYBot),
						.oXLeftPrev(drXLeftPrev),
						.oXRightPrev(drXRightPrev),
						.oYTopPrev(drYTopPrev),
						.oYBotPrev(drYBotPrev),
						.oXLeftPrev2(drXLeftPrev2),
						.oXRightPrev2(drXRightPrev2),
						.oYTopPrev2(drYTopPrev2),
						.oYBotPrev2(drYBotPrev2),
						.oXLeftPrev3(drXLeftPrev3),
						.oXRightPrev3(drXRightPrev3),
						.oYTopPrev3(drYTopPrev3),
						.oYBotPrev3(drYBotPrev3),
						.oRedArea(drRedArea),
						/////////////////////
						.iCLK(OSC_27),
						.iRST(!DLY2));

assign	dgCenterX = (dgXLeft < dgXRight) ? (((dgXRight - dgXLeft) >> 1) + dgXLeft) : 10'b0;
assign	dgCenterY = (dgYTop < dgYBot) ? (((dgYBot - dgYTop) >> 1) + dgYTop) : 10'b0;
												
assign	drCenterX = (drXLeft < drXRight) ? (((drXRight - drXLeft) >> 1) + drXLeft) : 10'b0;
assign	drCenterY = (drYTop < drYBot) ? (((drYBot - drYTop) >> 1) + drYTop) : 10'b0;

assign LED_GREEN = {KEY};
assign LED_RED = {DPDT_SW};


//	VGA Controller
VGA_Ctrl			u9 (// Host Side
						//.iRed(drRed),
						//.iGreen(drGreen),
						//.iBlue(drBlue),
						.iRed(red),
						.iGreen(green),
						.iBlue(blue),
						.oCurrent_X(VGA_X),
						.oCurrent_Y(VGA_Y),
						.oRequest(VGA_Read),
						// VGA Side
						.oVGA_R(VGA_R),
						.oVGA_G(VGA_G),
						.oVGA_B(VGA_B),
						.oVGA_HS(VGA_HS),
						.oVGA_VS(VGA_VS),
						.oVGA_SYNC(VGA_SYNC),
						.oVGA_BLANK(VGA_BLANK),
						.oVGA_CLOCK(VGA_CLK),
						// Control Signal
						.iCLK(OSC_27),
						.iRST_N(DLY2)
);
 					
/////////////////////////////////////
wire	[9:0]	drRed;
wire	[9:0]	drGreen;
wire	[9:0]	drBlue;
wire			drDVAL;
wire	[10:0]	drXLeft;
wire	[10:0]	drXRight;
wire	[10:0]	drYTop;
wire	[10:0]	drYBot;
wire	[10:0]	drXLeftPrev;
wire	[10:0]	drXRightPrev;
wire	[10:0]	drYTopPrev;
wire	[10:0]	drYBotPrev;
wire	[10:0]	drXLeftPrev2;
wire	[10:0]	drXRightPrev2;
wire	[10:0]	drYTopPrev2;
wire	[10:0]	drYBotPrev2;
wire	[10:0]	drXLeftPrev3;
wire	[10:0]	drXRightPrev3;
wire	[10:0]	drYTopPrev3;
wire	[10:0]	drYBotPrev3;
wire	[18:0]	drRedArea;

wire	[10:0]	drCenterX;
wire	[10:0]	drCenterY;
/////////////////////////////////////

/////////////////////////////////////
wire	[9:0]	dgRed;
wire	[9:0]	dgGreen;
wire	[9:0]	dgBlue;
wire			dgDVAL;
wire	[10:0]	dgXLeft;
wire	[10:0]	dgXRight;
wire	[10:0]	dgYTop;
wire	[10:0]	dgYBot;
wire	[10:0]	dgXLeftPrev;
wire	[10:0]	dgXRightPrev;
wire	[10:0]	dgYTopPrev;
wire	[10:0]	dgYBotPrev;
wire	[10:0]	dgXLeftPrev2;
wire	[10:0]	dgXRightPrev2;
wire	[10:0]	dgYTopPrev2;
wire	[10:0]	dgYBotPrev2;
wire	[10:0]	dgXLeftPrev3;
wire	[10:0]	dgXRightPrev3;
wire	[10:0]	dgYTopPrev3;
wire	[10:0]	dgYBotPrev3;
wire	[18:0]	dgRedArea;

wire	[10:0]	dgCenterX;
wire	[10:0]	dgCenterY;
/////////////////////////////////////


//	For ITU-R 656 Decoder
wire	[15:0]	YCbCr;
wire	[9:0]	TV_X;
wire			TV_DVAL;

//	For VGA Controller
wire	[9:0]	mRed;
wire	[9:0]	mGreen;
wire	[9:0]	mBlue;
wire	[10:0]	VGA_X;
wire	[10:0]	VGA_Y;
wire			VGA_Read;	//	VGA data request
wire			m1VGA_Read;	//	Read odd field
wire			m2VGA_Read;	//	Read even field

//	For YUV 4:2:2 to YUV 4:4:4
wire	[7:0]	mY;
wire	[7:0]	mCb;
wire	[7:0]	mCr;

//	For field select
wire	[15:0]	mYCbCr;
wire	[15:0]	mYCbCr_d;
wire	[15:0]	m1YCbCr;
wire	[15:0]	m2YCbCr;
wire	[15:0]	m3YCbCr;

//	For Delay Timer
wire			TD_Stable;
wire			DLY0;
wire			DLY1;
wire			DLY2;

//	For Down Sample
wire	[3:0]	Remain;
wire	[9:0]	Quotient;

assign	m1VGA_Read	=	VGA_Y[0]	?	1'b0		:	VGA_Read;
assign	m2VGA_Read	=	VGA_Y[0]	?	VGA_Read	:	1'b0	;
assign	mYCbCr_d	=	!VGA_Y[0]	?	m1YCbCr		:	m2YCbCr	;
assign	mYCbCr		=	m5YCbCr;

wire			mDVAL;

//	Line buffer, delay one line
Line_Buffer			u10(.clken(VGA_Read),
						.clock(OSC_27),
						.shiftin(mYCbCr_d),
						.shiftout(m3YCbCr)
);

Line_Buffer 		u11(.clken(VGA_Read),
						.clock(OSC_27),
						.shiftin(m3YCbCr),
						.shiftout(m4YCbCr)
);

wire	[15:0] m4YCbCr;
wire	[15:0] m5YCbCr;
wire	[8:0]	 Tmp1,Tmp2;
wire	[7:0]	 Tmp3,Tmp4;

assign Tmp1	   = m4YCbCr[7:0]+mYCbCr_d[7:0];
assign Tmp2	   = m4YCbCr[15:8]+mYCbCr_d[15:8];
assign Tmp3	   = Tmp1[8:2]+m3YCbCr[7:1];
assign Tmp4	   = Tmp2[8:2]+m3YCbCr[15:9];
assign m5YCbCr = {Tmp4,Tmp3};

AUDIO_DAC 			u12(// Audio Side
						.oAUD_BCK(AUD_BCLK),
						.oAUD_DATA(AUD_DACDAT),
						.oAUD_LRCK(AUD_DACLRCK),
						// Control Signals
						.iSrc_Select(2'b01),
						.iCLK_18_4(AUD_CTRL_CLK),
						.iRST_N(DLY1)
);

//==================================================================//

//PL
wire  	[9:0]	X_ADDRESS;					//	X coord on 7-seg display
wire  	[9:0]	Y_ADDRESS;					//	Y coord on 7-seg display
reg		[9:0]	CENTER_X_1;
reg		[9:0]	CENTER_Y_1;
reg		[9:0]	CENTER_X_2;
reg		[9:0]	CENTER_Y_2;
reg		[9:0] SHOT_CENTER_X;
reg		[9:0] SHOT_CENTER_Y;
reg				NEW_FRAME_1;
reg				NEW_FRAME_2;
reg signed [9:0] DIFF_X_1;
reg signed [9:0] DIFF_Y_1;
reg signed [9:0] DIFF_X_2;
reg signed [9:0] DIFF_Y_2;
reg signed [9:0] DIFF_CENTER_X;
reg signed [9:0] DIFF_CENTER_Y;
wire 				reset;
reg 		[3:0]	state;						//state machine

reg 		[2:0]		targ_counter;
reg 		[19:0] 	targ_array[4:0];
reg 		[9:0] 	targ_xcenter;
reg 		[9:0] 	targ_ycenter;

wire [10:0] targ_xright;
wire [10:0] targ_xleft;
wire [10:0] targ_yup;
wire [10:0] targ_ydown;
reg  [25:0] scounter;
reg  [10:0] seconds;
reg  [10:0] score;
reg new_shot;

assign targ_xright = {1'b0,targ_xcenter} + 11'd20;
assign targ_xleft  = {1'b0,targ_xcenter} - 11'd20;
assign targ_yup    = {1'b0,targ_ycenter} + 11'd20;
assign targ_ydown  = {1'b0,targ_ycenter} - 11'd20;

reg [9:0] red;
reg [9:0] blue;
reg [9:0] green;

//state names
parameter init=4'd0, test1=4'd1, test2=4'd2, test3=4'd3;

assign reset = ~KEY[0];

always@(posedge OSC_50) begin
	
	// VGA X, Y have wrapped around, set new frame flag.
	if (VGA_X == 20 && VGA_Y == 20) begin
		NEW_FRAME_1 <= 1'b1;
		NEW_FRAME_2 <= 1'b1;
	end
	
	if (reset) begin	//synch reset assumes KEY0 is held down 1/60 second
		//initialize variables
		scounter <= 26'd0;
		seconds  <= 11'd0;
      targ_xcenter <= 10'd0;
      targ_ycenter <= 10'd0;
		targ_array[0] <= {10'd320,10'd240};
		targ_array[1] <= {10'd320,10'd120};
		targ_array[2] <= {10'd480,10'd300};
		targ_array[3] <= {10'd120,10'd300};
		targ_array[4] <= {10'd480,10'd120};
		targ_counter <= 3'd0;
		
		CENTER_X_1 <= 10'd20;
		CENTER_Y_1 <= 10'd20;
		CENTER_X_2 <= 10'd620;
		CENTER_Y_2 <= 10'd20;
		SHOT_CENTER_X <= 10'h0;
		SHOT_CENTER_Y <= 10'h0;
		score 	<= 0;
		new_shot <= 0;
		state		<= init;	//first state in regular state machine
	end
	else if(~KEY[1]) begin // shot detected
		SHOT_CENTER_X <= drCenterX;
		SHOT_CENTER_Y <= drCenterY;
		new_shot = 1;
	end
	//modify display during sync
	else if (VGA_VS | VGA_HS) begin

		scounter <= (scounter != 26'd50000000) ? (scounter + 26'd1): 26'd0;
		seconds <= (scounter == 26'd50000000) ? ((seconds == DPDT_SW[3:1]) ? 0: seconds+1):seconds;
		
		targ_counter <= (scounter == 26'd50000000 && seconds == DPDT_SW[3:1])? ((targ_counter == 4) ? 0 : targ_counter+1):targ_counter;
      targ_xcenter <= targ_array[targ_counter][19:10];
      targ_ycenter <= targ_array[targ_counter][9:0];
		
		if (new_shot && (SHOT_CENTER_X >= targ_xleft && SHOT_CENTER_X <= targ_xright &&
		    SHOT_CENTER_Y >= targ_ydown && SHOT_CENTER_Y <= targ_yup)) begin
			 new_shot = 0;
			 score <= score + 1;
		end
		
		case(state)

			init: begin
				state <= test1 ;
			end

			test1: begin //locate center
				// player 1
				if (drRed == 10'hFFFF) begin
					if (NEW_FRAME_1 == 1 &&
					    VGA_X >= 32      && // avoid left border
					    VGA_X < 608      && // avoid right border
					    VGA_Y >= 112     && // avoid top border
					    VGA_Y < 388)        // avoid bottom border
					begin
						CENTER_X_1 <= drCenterX;
						CENTER_Y_1 <= drCenterY;
						NEW_FRAME_1 <= 1'b0;
					end
				end
				/* player 2
				//if (drBlue == 10'hFFFF) begin
					if (NEW_FRAME_2 == 1 &&
					    VGA_X >= 32      && // avoid left border
					    VGA_X <  608     && // avoid right border
					    VGA_Y >= 112     && // avoid top border
					    VGA_Y <  388)       // avoid bottom border
					begin
						CENTER_X_2 <= dgCenterX;
						CENTER_Y_2 <= dgCenterY;
						NEW_FRAME_2 <= 1'b0;
					end
				//end*/
				state <= test2 ;
			end

			test2: begin //calculate diff
				DIFF_X_1 <= VGA_X - CENTER_X_1;
				DIFF_Y_1 <= VGA_Y - CENTER_Y_1;
				DIFF_X_2 <= VGA_X - CENTER_X_2;
				DIFF_Y_2 <= VGA_Y - CENTER_Y_2;
				DIFF_CENTER_X <= VGA_X - SHOT_CENTER_X;
				DIFF_CENTER_Y <= VGA_Y - SHOT_CENTER_Y;
				state <= test3 ;
			end

			test3: begin
				// draw puck
				if ((SHOT_CENTER_X != 0 && SHOT_CENTER_Y != 0) &&
				   (( DIFF_CENTER_X <= 3  && // the square
				      DIFF_CENTER_X >= -3 &&
				      DIFF_CENTER_Y <= 3  &&
				      DIFF_CENTER_Y >= -3)
				    ||
				    /*( DIFF_CENTER_X == 9  && // the right arc
				      DIFF_CENTER_Y <= 6  &&
				      DIFF_CENTER_Y >= -6)
				    ||
				    ( DIFF_CENTER_X == 10 &&
				      DIFF_CENTER_Y <= 5  &&
				      DIFF_CENTER_Y >= -5)
				    ||
				    ( DIFF_CENTER_X == 11 &&
				      DIFF_CENTER_Y <= 3  &&
				      DIFF_CENTER_Y >= -3)
				    ||*/
				    ( DIFF_CENTER_X <= 12 &&
				      DIFF_CENTER_Y == 0 )
				    ||
				    /*( DIFF_CENTER_X == -9 && // the left arc
				      DIFF_CENTER_Y <= 6  &&
				      DIFF_CENTER_Y >= -6)
				    ||
				    ( DIFF_CENTER_X == -10 &&
				      DIFF_CENTER_Y <= 5   &&
				      DIFF_CENTER_Y >= -5)
				    ||
				    ( DIFF_CENTER_X == -11 &&
				      DIFF_CENTER_Y <= 3   &&
				      DIFF_CENTER_Y >= -3)
				    ||*/
				    ( DIFF_CENTER_X >= -12 &&
				      DIFF_CENTER_Y == 0 )
				    ||
				    /*( DIFF_CENTER_Y == 9 && // the top arc
				      DIFF_CENTER_X <= 6 &&
				      DIFF_CENTER_X >= -6)
				    ||
				    ( DIFF_CENTER_Y == 10 &&
				      DIFF_CENTER_X <= 5  &&
				      DIFF_CENTER_X >= -5)
				    ||
				    ( DIFF_CENTER_Y == 11 &&
				      DIFF_CENTER_X <= 3  &&
				      DIFF_CENTER_X >= -3)
				    ||*/
				    ( DIFF_CENTER_Y <= 12 &&
				      DIFF_CENTER_X == 0 )
				    ||
				    /*( DIFF_CENTER_Y == -9 && // the bottom arc
				      DIFF_CENTER_X <= 6  &&
				      DIFF_CENTER_X >= -6)
				    ||
				    ( DIFF_CENTER_Y == -10 &&
				      DIFF_CENTER_X <= 5   &&
				      DIFF_CENTER_X >= -5)
				    ||
				    ( DIFF_CENTER_Y == -11 &&
				      DIFF_CENTER_X <= 3   &&
				      DIFF_CENTER_X >= -3)
				    ||*/
				    ( DIFF_CENTER_Y >= -12 &&
				      DIFF_CENTER_X == 0 ))
				) begin
					green	<=	10'b1010100000;
					red	<=	10'b1010100000;
					blue	<=	10'b0000000000;
				end
				else begin
					// player 1
					if ((CENTER_X_1 != 0 && CENTER_Y_1 != 0) &&
						(( 	DIFF_X_1 <= 8  && // the square
							DIFF_X_1 >= -8 &&
							DIFF_Y_1 <= 8  &&
							DIFF_Y_1 >= -8)
					    ||
					    (	DIFF_X_1 == 9  && // the right arc
							DIFF_Y_1 <= 6  &&
							DIFF_Y_1 >= -6)
					    ||
					    ( 	DIFF_X_1 == 10 &&
							DIFF_Y_1 <= 5  &&
							DIFF_Y_1 >= -5)
					    ||
					    ( 	DIFF_X_1 == 11 &&
							DIFF_Y_1 <= 3  &&
							DIFF_Y_1 >= -3)
					    ||
					    ( 	DIFF_X_1 == 12 &&
							DIFF_Y_1 == 0 )
					    ||
					    ( 	DIFF_X_1 == -9 && // the left arc
							DIFF_Y_1 <= 6  &&
							DIFF_Y_1 >= -6)
					    ||
					    ( 	DIFF_X_1 == -10 &&
							DIFF_Y_1 <= 5   &&
							DIFF_Y_1 >= -5)
					    ||
					    ( 	DIFF_X_1 == -11 &&
							DIFF_Y_1 <= 3   &&
							DIFF_Y_1 >= -3)
					    ||
					    ( 	DIFF_X_1 == -12 &&
							DIFF_Y_1 == 0 )
					    ||
					    ( 	DIFF_Y_1 == 9 && // the top arc
							DIFF_X_1 <= 6 &&
							DIFF_X_1 >= -6)
					    ||
					    ( 	DIFF_Y_1 == 10 &&
							DIFF_X_1 <= 5  &&
							DIFF_X_1 >= -5)
					    ||
					    ( 	DIFF_Y_1 == 11 &&
							DIFF_X_1 <= 3  &&
							DIFF_X_1 >= -3)
					    ||
					    ( 	DIFF_Y_1 == 12 &&
							DIFF_X_1 == 0 )
					    ||
					    ( 	DIFF_Y_1 == -9 && // the bottom arc
							DIFF_X_1 <= 6  &&
							DIFF_X_1 >= -6)
					    ||
					    ( 	DIFF_Y_1 == -10 &&
							DIFF_X_1 <= 5   &&
							DIFF_X_1 >= -5)
					    ||
					    ( 	DIFF_Y_1 == -11 &&
							DIFF_X_1 <= 3   &&
							DIFF_X_1 >= -3)
					    ||
					    ( 	DIFF_Y_1 == -12 &&
							DIFF_X_1 == 0 ))
					)
					begin	// blue player
						green	<=	10'b1010100000;
						red		<=	10'b0000000000;
						blue	<=	10'b1010100000;
					end
					// player 2
					else if ((CENTER_X_2 != 0 && CENTER_Y_2 != 0) &&
						(( 	DIFF_X_2 <= 8  && // the square
							DIFF_X_2 >= -8 &&
							DIFF_Y_2 <= 8  &&
							DIFF_Y_2 >= -8)
					    ||
					    ( 	DIFF_X_2 == 9  && // the right arc
							DIFF_Y_2 <= 6  &&
							DIFF_Y_2 >= -6)
					    ||
					    ( 	DIFF_X_2 == 10 &&
							DIFF_Y_2 <= 5  &&
							DIFF_Y_2 >= -5)
					    ||
					    ( 	DIFF_X_2 == 11 &&
							DIFF_Y_2 <= 3  &&
							DIFF_Y_2 >= -3)
					    ||
					    ( 	DIFF_X_2 == 12 &&
							DIFF_Y_2 == 0 )
					    ||
					    ( 	DIFF_X_2 == -9 && // the left arc
							DIFF_Y_2 <= 6  &&
							DIFF_Y_2 >= -6)
					    ||
					    ( 	DIFF_X_2 == -10 &&
							DIFF_Y_2 <= 5   &&
							DIFF_Y_2 >= -5)
					    ||
					    ( 	DIFF_X_2 == -11 &&
							DIFF_Y_2 <= 3   &&
							DIFF_Y_2 >= -3)
					    ||
					    ( 	DIFF_X_2 == -12 &&
							DIFF_Y_2 == 0 )
					    ||
					    ( 	DIFF_Y_2 == 9 && // the top arc
							DIFF_X_2 <= 6 &&
							DIFF_X_2 >= -6)
					    ||
					    ( 	DIFF_Y_2 == 10 &&
							DIFF_X_2 <= 5  &&
							DIFF_X_2 >= -5)
					    ||
					    ( 	DIFF_Y_2 == 11 &&
							DIFF_X_2 <= 3  &&
							DIFF_X_2 >= -3)
					    ||
					    ( 	DIFF_Y_2 == 12 &&
							DIFF_X_2 == 0 )
					    ||
					    ( 	DIFF_Y_2 == -9 && // the bottom arc
							DIFF_X_2 <= 6  &&
							DIFF_X_2 >= -6)
					    ||
					    ( 	DIFF_Y_2 == -10 &&
							DIFF_X_2 <= 5   &&
							DIFF_X_2 >= -5)
					    ||
					    ( 	DIFF_Y_2 == -11 &&
							DIFF_X_2 <= 3   &&
							DIFF_X_2 >= -3)
					    ||
					    ( 	DIFF_Y_2 == -12 &&
							DIFF_X_2 == 0 ))
					)
					begin	// red player
						green	<=	10'b0000000000;
						red		<=	10'b1010100000;
						blue	<=	10'b1010100000;
					end
					// draw left border
					else if ((VGA_X == targ_xleft || VGA_X == (targ_xleft-1))   &&
							 (VGA_Y >= targ_ydown && VGA_Y < targ_yup))
					begin
						green	<=	10'b0000000000;
						red	<=	10'b0000000000;
						blue	<=	10'b1010100000;
					end
					// draw right border
					else if ((VGA_X == targ_xright || VGA_X == (targ_xright-1))   &&
							 (VGA_Y >= targ_ydown && VGA_Y < targ_yup))
					begin
						green	<=	10'b0000000000;
						red	<=	10'b0000000000;
						blue	<=	10'b1010100000;
					end
					// draw top border
					else if ((VGA_X >= targ_xleft && VGA_X < targ_xright)   &&
							 (VGA_Y == targ_yup || VGA_Y == (targ_yup-1)))
					begin
						green	<=	10'b0000000000;
						red	<=	10'b0000000000;
						blue	<=	10'b1010100000;
					end
					// draw down border
					else if ((VGA_X >= targ_xleft && VGA_X < targ_xright)   &&
							 (VGA_Y == targ_ydown || VGA_Y == (targ_ydown-1)))
					begin
						green	<=	10'b0000000000;
						red	<=	10'b0000000000;
						blue	<=	10'b1010100000;
					end
					// blacken all else
					else if (VGA_X >= 20 &&
					         VGA_X < 640)
					begin
						if(DPDT_SW[0]) begin
							red	<=	10'b0000000000;
							blue	<=	10'b0000000000;
							green	<=	10'b0000000000;
						end else if(DPDT_SW[17]) begin
							red	<=	drRed;
							blue	<=	drBlue;
							green	<=	drGreen;
						end
						else begin
							red	<=	mRed;
							blue	<=	mBlue;
							green	<=	mGreen;
						end
					end
				end
				state <= test1 ;
			end
		endcase
	end

end
					
endmodule