
module green_block (
	input					clk,
	input					reset,
	input		[10:0]	i_X,
	input		[10:0]	i_Y,
	input		[9:0]		i_R,
	input		[9:0]		i_G,
	input		[9:0]		i_B,
	output	[9:0]		o_avg_R,
	output	[9:0]		o_avg_G,
	output	[9:0]		o_avg_B,
	output				o_Block,
	output				o_inBlock
);

localparam Pos_X = 320;
localparam Pos_Y = 240;
localparam HalfWidth = 20;
localparam HalfHeight = 20;

wire X_equ = ( i_X == Pos_X - HalfWidth + 1 ) || ( i_X == Pos_X + HalfWidth  );
wire Y_equ = ( i_Y == Pos_Y - HalfHeight + 1) || ( i_Y == Pos_Y + HalfHeight );
wire X_in  = ( i_X > Pos_X - HalfWidth   ) && ( i_X <= Pos_X + HalfWidth  );
wire Y_in  = ( i_Y > Pos_Y - HalfHeight  ) && ( i_Y <= Pos_Y + HalfHeight );

wire out_of_range = ( i_X > Pos_X + HalfWidth ) && ( i_Y > Pos_Y + HalfHeight );

assign o_Block = X_equ ? ( Y_in ? 1'b1 : 1'b0 ) 
               : Y_equ ? ( X_in ? 1'b1 : 1'b0 )
					: 1'b0;

assign o_inBlock = X_in && Y_in;

reg [19:0] sum_R;
reg [19:0] sum_G;
reg [19:0] sum_B;
reg in_blk_stats;

reg [9:0] avg_R;
reg [9:0] avg_G;
reg [9:0] avg_B;

reg [10:0] pre_X;
reg [10:0] pre_Y;

assign o_avg_R = avg_R;
assign o_avg_G = avg_G;
assign o_avg_B = avg_B;

always @( posedge clk )
begin
	if ( reset )
	begin
		pre_X <= 11'b0;
		pre_Y <= 11'b0;
		sum_R <= 20'b0;
		sum_G <= 20'b0;
		sum_B <= 20'b0;
		avg_R <= 10'd1023;
		avg_G <= 10'd1023;
		avg_B <= 10'd1023;
		in_blk_stats <= 1'b0;
	end
	else
	begin
		if ( ( pre_X != i_X ) || ( pre_Y != i_Y ))
		begin
			pre_X <= i_X;
			pre_Y <= i_Y;
			if ( o_inBlock && in_blk_stats )
			begin
				sum_R <= sum_R + { 10'b0, i_R };
				sum_G <= sum_G + { 10'b0, i_G };
				sum_B <= sum_B + { 10'b0, i_B };
				in_blk_stats <= 1'b1;
			end
			else
				if ( o_inBlock && !in_blk_stats )
				begin
					sum_R <= i_R;
					sum_G <= i_G;
					sum_B <= i_B;
					in_blk_stats <= 1'b1;
				end
				else
					if ( out_of_range && in_blk_stats )
					begin
						avg_R <= sum_R[17:8];
						avg_G <= sum_G[17:8];
						avg_B <= sum_B[17:8];
						in_blk_stats <= 1'b0;
					end
		end
	end
end

endmodule