module neuralNet
	(
		CLK, 
		RESET, 
		boxXCenter,
		oleftInNeuOut,
		orightInNeuOut,
		oleftOutNeuOut,
		orightOutNeuOut,
		oleftInNeuSpike,
		orightInNeuSpike,
		oleftOutNeuSpike,
		orightOutNeuSpike
	);
	
	input				CLK;
	input				RESET;
	input		[9:0]	boxXCenter;
	
	output		[17:0]	oleftInNeuOut, orightInNeuOut;
	output		[17:0]	oleftOutNeuOut, orightOutNeuOut;
	output				oleftInNeuSpike, orightInNeuSpike;
	output				oleftOutNeuSpike, orightOutNeuSpike;
	
	wire signed	[11:0]	paddedBoxXCenter;
	wire signed	[10:0]	leftDist, rightDist;
	wire signed	[17:0]	leftInNeuIn, rightInNeuIn;
	wire signed	[17:0]	leftInNeuOut, rightInNeuOut;
	wire signed	[17:0]	leftOutNeuIn, rightOutNeuIn;
	wire signed	[17:0]	leftOutNeuOut, rightOutNeuOut;
	wire				leftInNeuSpike, rightInNeuSpike;
	wire				leftOutNeuSpike, rightOutNeuSpike;
	
	assign oleftInNeuOut		=	leftInNeuOut;
	assign orightInNeuOut		=	rightInNeuOut;
	assign oleftOutNeuOut		=	leftOutNeuOut;
	assign orightOutNeuOut		=	rightOutNeuOut;
	assign oleftInNeuSpike		=	leftInNeuSpike;
	assign orightInNeuSpike		=	rightInNeuSpike;
	assign oleftOutNeuSpike		=	leftInNeuSpike;
	assign orightOutNeuSpike	=	rightOutNeuSpike;
	
	wire 		[3:0]	a1, b, a2;
	wire signed	[17:0]	c, d;					// parameters
	
	//burster parameters
	assign a1 		=	6;						// 0.016
	assign b 		=	2;						// 0.25
	// SCALE c, d, and I by 0.01 from Izhikevich table values
	assign c		= 	18'sh3_8000;			// -0.5 Izhikevich table value (50) times 0.01
	assign d		= 	18'sh0_051E;			// 0.02 Izhikevich table value (2) times 0.01
	assign a2		=	6;
	
	//find signed left and right distances of the
	//x-center of bounding box to center of screen
	//***THESE CAN BE WRONG! CHECK TO MAKE SURE IN CORRECT FIXED POINT FORMAT!!!***//
	assign paddedBoxXCenter	=	{2'sb0, boxXCenter};
	//assign leftDist		=	(18'sd350 - paddedBoxXCenter);
	//assign rightDist	=	(paddedBoxXCenter - 18'sd290);
	assign leftInNeuIn	=	/*(paddedBoxXCenter <= 13'sd350) ? */{(12'sd350 - paddedBoxXCenter),6'b0};// : 18'b0;
	assign rightInNeuIn	=	/*(paddedBoxXCenter >= 13'sd290) ? */{(paddedBoxXCenter - 12'sd290),6'b0};// : 18'b0;
	
	//Two input chattering burster neurons, not connected to each other
	//Iz_neuron(out,spike,a,b,c,d,I,clk,reset);
	//do I want to zero out the input when center on other side, or keep using signed and neg vals
	//need to look more into these a,b,c,d values
	Iz_neuron	leftInputNeu	(	.out(leftInNeuOut),
									.spike(leftInNeuSpike),
									.a(a1),
									.b(b),
									.c(c),
									.d(d),
									.I(leftInNeuIn),
									.clk(CLK),
									.reset(RESET)
								);
	Iz_neuron	rightInputNeu	(	.out(rightInNeuOut),
									.spike(rightInNeuSpike),
									.a(a2),
									.b(b),
									.c(c),
									.d(d),
									.I(rightInNeuIn),
									.clk(CLK),
									.reset(RESET)
								);
		
	//create excitory synapses for corresponding sides - (single_synapse?)
	//create inhibitory synapses for opposite sides (make them weaker) - (single_synapse?)
	//figure out how to combine them (electrical_synapse? single_synapse outputs as inputs?)
	//multi_synapse(out,s1,w1,s2,w2,tau,in1,in2,clk,reset);
	//synapse(out,s1,w1,s2,w2,s3,w3,clk,reset);
	synapse			leftOutSyn	(	.out(leftOutNeuIn),
									.s1(leftInNeuSpike),
									.w1(18'sh0_f800),
									.s2(rightInNeuSpike),
									.w2(18'sh3_f000),
									.s3(1'b0),
									.w3(18'b0),
									.clk(CLK),
									.reset(RESET)
								);
	synapse			rightOutSyn	(	.out(rightOutNeuIn),
									.s1(leftInNeuSpike),
									.w1(18'sh3_f000),
									.s2(rightInNeuSpike),
									.w2(18'sh0_f800),
									.s3(1'b0),
									.w3(18'b0),
									.clk(CLK),
									.reset(RESET)
								);
	/*multi_synapse	leftOutSyn	(	.out(leftOutNeuIn),
									.s1(leftInNeuSpike),
									.w1(18'sh0_f800),
									.s2(rightInNeuSpike),
									.w2(18'sh3_f000),
									.tau(4'd4),
									.in1(leftInNeuOut),
									.in2(rightInNeuOut),
									.clk(CLK),
									.reset(RESET)
								);
	multi_synapse	rightOutSyn	(	.out(rightOutNeuIn),
									.s1(rightInNeuSpike),
									.w1(18'sh0_f800),
									.s2(leftInNeuSpike),
									.w2(18'sh3_f000),
									.tau(4'd4),
									.in1(rightInNeuOut),
									.in2(leftInNeuOut),
									.clk(CLK),
									.reset(RESET)
								);*/
	
	//Two output chattering burster neurons, not connected to each other
	//Iz_neuron(out,spike,a,b,c,d,I,clk,reset);
	Iz_neuron	leftOutputNeu	(	.out(leftOutNeuOut),
									.spike(leftOutNeuSpike),
									.a(a1),
									.b(b),
									.c(c),
									.d(d),
									.I(leftOutNeuIn),
									.clk(CLK),
									.reset(RESET)
								);
	Iz_neuron	rightOutputNeu	(	.out(rightOutNeuOut),
									.spike(rightOutNeuSpike),
									.a(a2),
									.b(b),
									.c(c),
									.d(d),
									.I(rightOutNeuIn),
									.clk(CLK),
									.reset(RESET)
								);
	/*Iz_neuron	rightOutputNeu	(	.out(rightOutNeuOut),
									.spike(rightOutNeuSpike),
									.a(a2),
									.b(b),
									.c(c),
									.d(d),
									.I(rightOutNeuIn),
									.clk(CLK),
									.reset(RESET)
								);*/

endmodule

//////////////////////////////////////////////////
//// Izhikevich neuron ///////////////////////////
//////////////////////////////////////////////////
// Modified to eliminate two of the multiplies
// Since a nd b need not be very precise 
// and because 0.02<a<0.2 and 0.1<b<0.3
// Treat the a and b inputs as an index do determine how much to 
// shift the intermediate results
// treat the a input as value>>>a with a=1 to 7
// treat the b input as value>>>b with b=1 to 3
// NOTE: constants c,d, and I as given in Izhikevich's
// tables and figures have to be multiplied by 0.01
module Iz_neuron(out,spike,a,b,c,d,I,clk,reset);
	output			[17:0]	out;		// the simulated membrane voltage
	output					spike;		// the action potential output
	input	signed	[17:0]	c, d, I;	// a,b,c,d set cell dynamics, I is the input current
	input 			[3:0]	a, b;
	input 					clk, reset;
	
	wire	signed	[17:0]	out;
	reg 					spike;
	reg		signed	[17:0]	v1,u1;
	wire	signed	[17:0]	u1reset, v1new, u1new, du1;
	wire	signed	[17:0]	v1xv1, v1xb;
	wire	signed	[17:0]	p, c14;
	
	assign p 	=	18'sh0_4CCC;		// 0.30
	assign c14 	=	18'sh1_6666;		// 1.4
	
	assign out	=	v1;  //copy state var to output
	
	always @ (posedge clk)
	begin
		if (reset)
		begin	
			v1		<=	18'sh3_4CCD;	// -0.7
			u1		<=	18'sh3_CCCD;	// -0.2
			spike	<=	0;
		end	
		else 
		begin
			if ((v1 > p)) 
			begin 
				v1		<=	c ; 		
				u1		<=	u1reset ;
				spike	<=	1;
			end
			else
			begin
				v1		<=	v1new ;
				u1		<=	u1new ; 
				spike	<=	0;	
			end
		end 
	end
	
	// dt = 1/16 or 2>>4
	// v1(n+1) = v1(n) + dt*(4*(v1(n)^2) + 5*v1(n) +1.40 - u1(n) + I)
	// but note that what is actually coded is
	// v1(n+1) = v1(n) + (v1(n)^2) + 5/4*v1(n) +1.40/4 - u1(n)/4 + I/4)/4
	signed_mult v1sq(v1xv1, v1, v1);
	assign v1new	=	v1 + ((v1xv1 + v1+(v1>>>2) + (c14>>>2) - (u1>>>2) + (I>>>2))>>>2);
	
	// u1(n+1) = u1 + dt*a*(b*v1(n) - u1(n))
	assign v1xb		=	v1>>>b;         //mult (v1xb, v1, b);
	assign du1		=	(v1xb-u1)>>>a ;  //mult (du1, (v1xb-u1), a);
	assign u1new	=	u1 + (du1>>>4) ; 
	assign u1reset	=	u1 + d ;
endmodule
//////////////////////////////////////////////////

////////////////////////////////////////////
//// Electrical Synapse ////////////////////
////////////////////////////////////////////
// Acts to create a resistance coupling two cells 
// I (into cell 2) = (v1 - v2)*conductance
// I (into cell 1) = (v2 - v1)*conductance
// conductance is 0<conductance<15 which is a right-shift factor
// rect specifies: 
//      0=no rectification, 
//      1=current from 1 to 2
//      2=current from 2 to 1
// the 'in' allows synapes to be daisy-chained
////////////////////////////////////////
module electrical_synapse(I2_out, I1_out, v2, v1, cond, rect, in2, in1);
	output	reg	signed	[17:0]	I2_out, I1_out; 	//the simulated synaptic current
	input		signed	[17:0]	in1, in2;			// the input currents
	input				[3:0]	cond, rect;			// conductance and rectifcation
	input		signed	[17:0]	v1, v2; 
	  		
	wire	signed		[17:0]	v_diff;
	reg		signed		[17:0]	rect_v_diff ;
	
	assign v_diff	=	v1 - v2 ;
	always @(*)
	begin
		case (rect)
			0:	rect_v_diff	<=	v_diff ;  
			1:	rect_v_diff	<=	v_diff>0? v_diff : 18'd0; 
			2:	rect_v_diff	<=	v_diff<0? v_diff : 18'd0;  
		endcase
		I2_out	<=	((rect_v_diff)>>>cond) + in2;
		I1_out	<=	((-rect_v_diff)>>>cond) + in1;
	end
endmodule

////////////////////////////////////////
//// Single Synapse ////////////////////
////////////////////////////////////////
// Acts to create an exponentially falling current at the output 
// from a spike input and a weight which can be +/-
// One spike input and weight are defined, along with an input value 
// so that synapes can be daisy-chained.
// Tau is 2<tau<8 which sets the decay time constant.
// Tau=2 is approx 0.25 mSec, tau=4 is approx 1.0 mSec, tau=6 is 4 mSec.
////////////////////////////////////////
module single_synapse(out,s1,w1,tau,in,clk,reset);
	output	signed	[17:0]	out;		//the simulated synaptic current
	input	signed	[17:0]	in;			// the input current
	input			[3:0]	tau;		// the synaptic current decay time const
	input					s1;     	// the action potential inputs
	input	signed	[17:0]	w1;   		// synaptic weight
	input					clk, reset;
	
	wire	signed	[17:0]	out;
	reg		signed	[17:0]	v1;
	wire	signed	[17:0]	v1new;
	
	always @ (posedge clk) 
	begin
		if (reset)	v1	<=	18'sh0; // no synaptic input at t=0
		else		v1	<=	v1new;
	end
	
	// dt = 1/16 or 2>>4 and tau=1
	// v1(n+1) = v1(n) + dt/tau*(-v1(n)+s1*w1+s2*w2+s3*w3)
	assign v1new	=	v1 + ((-v1)>>>tau) + (s1?w1:0);
	assign out		=	v1 + in ;						//copy state var to output & add input
	
endmodule


//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////
//// Multi Synapse ////////////////////
////////////////////////////////////////
// Acts to create an exponentially falling current at the output 
// from a spike input and a weight which can be +/-
// One spike input and weight are defined, along with an input value 
// so that synapes can be daisy-chained.
// Tau is 2<tau<8 which sets the decay time constant.
// Tau=2 is approx 0.25 mSec, tau=4 is approx 1.0 mSec, tau=6 is 4 mSec.
////////////////////////////////////////
module multi_synapse(out,s1,w1,s2,w2,tau,in1,in2,clk,reset);
	output	signed	[17:0]	out;		//the simulated synaptic current
	input	signed	[17:0]	in1, in2;	// the input current
	input			[3:0]	tau;		// the synaptic current decay time const
	input					s1, s2;    	// the action potential inputs
	input	signed	[17:0]	w1, w2;		// synaptic weight
	input					clk, reset;
	
	wire	signed	[17:0]	out;
	reg		signed	[17:0]	v1;
	wire	signed	[17:0]	v1new;
	
	always @ (posedge clk) 
	begin
		if (reset)	v1	<=	18'sh0; // no synaptic input at t=0
		else		v1	<=	v1new;
	end
	
	// dt = 1/16 or 2>>4 and tau=1
	// v1(n+1) = v1(n) + dt/tau*(-v1(n)+s1*w1+s2*w2+s3*w3)
	assign v1new	=	v1 + ((-v1)>>>tau) + (s1?w1:0) + (s2?w2:0);
	assign out		=	v1 + in1 + in2 ;				//copy state var to output & add input
	
endmodule
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////
//// Synapse ///////////////////////////
////////////////////////////////////////
// Acts to create an exponentially falling current at the output 
// from a spike input and a weight which can be +/-
// up to three spike inputs are defined, each with its own weight
module synapse(out,s1,w1,s2,w2,s3,w3,clk,reset);
	output [17:0] out; 				//the simulated current
	input s1,s2,s3;     			// the action potential inputs
	input signed [17:0] w1,w2,w3;   //weights
	input clk, reset;
	
	wire signed	[17:0] out;
	reg  signed	[17:0] v1 ;
	wire signed	[17:0] v1new ;
	
	assign out = v1;  //copy state var to output
	
	always @ (posedge clk) 
	begin
		if (reset) v1 <= 18'sh0; //
		else v1 <= v1new ;		 
	end
	
	// dt = 1/16 or 2>>4 and tau=1
	// v1(n+1) = v1(n) + dt/tau*(-v1(n)+s1*w1+s2*w2+s3*w3)
	assign v1new = v1 + ((-v1)>>>4) + (s1?w1:0)+(s2?w2:0)+(s3?w3:0) ;
	
endmodule

//////////////////////////////////////////////////
//// signed mult of 2.16 format 2'comp////////////
//////////////////////////////////////////////////
module signed_mult (out, a, b);

	output			[17:0]	out;
	input 	signed	[17:0]	a;
	input 	signed	[17:0]	b;
	
	wire	signed	[17:0]	out;
	wire 	signed	[35:0]	mult_out;

	assign mult_out	=	a * b;
	//assign out = mult_out[33:17];
	assign out		=	{mult_out[35], mult_out[32:16]};
endmodule

//////////////////////////////////////////////////
/*
%cell parameters from Izhikevich web page
pars=[0.02      0.2     -65      6       14 ;...    % tonic spiking
    0.02      0.25    -65      6       0.5 ;...   % phasic spiking
    0.02      0.2     -50      2       15 ;...    % tonic bursting
    0.02      0.25    -55     0.05     0.6 ;...   % phasic bursting
    0.02      0.2     -55     4        10 ;...    % mixed mode
    0.01      0.2     -65     8        30 ;...    % spike frequency adaptation
    0.02      -0.1    -55     6        0  ;...    % Class 1
    0.2       0.26    -65     0        0  ;...    % Class 2
    0.02      0.2     -65     6        7  ;...    % spike latency
    0.05      0.26    -60     0        0  ;...    % subthreshold oscillations
    0.1       0.26    -60     -1       0  ;...    % resonator
    0.02      -0.1    -55     6        0  ;...    % integrator
    0.03      0.25    -60     4        0;...      % rebound spike
    0.03      0.25    -52     0        0;...      % rebound burst
    0.03      0.25    -60     4        0  ;...    % threshold variability
    1         1.5     -60     0      -65  ;...    % bistability
    1       0.2     -60     -21      0  ;...    % DAP
    0.02      1       -55     4        0  ;...    % accomodation
    -0.02      -1      -60     8        80 ;...    % inhibition-induced spiking
    -0.026     -1      -45     0        80];       % inhibition-induced bursting
*/
//////////////////////////////////////////////////////